// standard includes
#include <assert.h>
#include <iostream>
#include <memory>

// subproject includes
#include <tao/pegtl.hpp>

// includes
#include "csg_types.hpp"

using namespace tao::pegtl;

namespace {

struct parser_state {
  std::vector<std::vector<csg::Type>> current_objs;
  std::vector<csg::Type> current_group;
  std::string current_name;
  std::vector<double> current_vec;
  std::vector<std::vector<double>> current_matrix;
  std::vector<std::vector<std::vector<double>>> current_matrices;
  std::map<std::string,
           std::variant<double, bool, std::vector<double>, std::string>>
      curr_attr;
};

} // namespace

namespace {

struct escaped_x : seq<one<'x'>, rep<2, must<xdigit>>> {};
struct escaped_u : seq<one<'u'>, rep<4, must<xdigit>>> {};
struct escaped_U : seq<one<'U'>, rep<8, must<xdigit>>> {};
struct escaped_c
    : one<'\'', '"', '?', '\\', 'a', 'b', 'f', 'n', 'r', 't', 'v'> {};

struct escaped : sor<escaped_x, escaped_u, escaped_U, escaped_c> {};

struct character
    : if_must_else<one<'\\'>, escaped, utf8::range<0x20, 0x10FFFF>> {};
struct string_literal : if_must<one<'"'>, until<one<'"'>, character>> {};

struct plus_minus : opt<one<'+', '-'>> {};
struct dot : one<'.'> {};

struct decimal
    : if_then_else<dot, plus<digit>, seq<plus<digit>, opt<dot, star<digit>>>> {
};
struct e : one<'e', 'E'> {};
struct p : one<'p', 'P'> {};
struct exponent : seq<plus_minus, plus<digit>> {};
struct double_ : seq<plus_minus, decimal, opt<e, exponent>> {};

struct padded_double : pad<double_, space> {};

struct L_FUN : pad<string<'('>, space> {};
struct R_FUN : pad<string<')'>, space> {};
struct L_ARR : pad<string<'['>, space> {};
struct R_ARR : pad<string<']'>, space> {};
struct L_BLK : pad<string<'{'>, space> {};
struct R_BLK : pad<string<'}'>, space> {};
struct S_CLN : pad<string<';'>, space> {};

struct true_literal : string<'t', 'r', 'u', 'e'> {};
struct false_literal : string<'f', 'a', 'l', 's', 'e'> {};

struct boolean_literal : sor<true_literal, false_literal> {};

struct special_identifier : seq<string<'$'>, identifier> {};

struct name : sor<identifier, special_identifier> {};

struct vector_cell : padded_double {};

struct vector : seq<L_ARR, list<vector_cell, one<','>>, R_ARR> {};

struct vector_attr : vector {};

struct value
    : sor<padded_double, boolean_literal, string_literal, vector_attr> {};

struct keyval : seq<pad<name, space>, string<'='>, pad<value, space>> {};

struct attr_list : list<keyval, one<','>> {};

struct row : vector {};

struct matrix : seq<L_ARR, list<row, one<','>>, R_ARR> {};

struct sphere
    : seq<string<'s', 'p', 'h', 'e', 'r', 'e'>, L_FUN, attr_list, R_FUN> {};

struct cube : seq<string<'c', 'u', 'b', 'e'>, L_FUN, attr_list, R_FUN> {};

struct cylinder : seq<string<'c', 'y', 'l', 'i', 'n', 'd', 'e', 'r'>, L_FUN,
                      attr_list, R_FUN> {};

struct shape : seq<sor<cube, cylinder, sphere>, opt<S_CLN>> {};

struct obj_list;

struct bool_union : seq<string<'u', 'n', 'i', 'o', 'n'>, L_FUN, R_FUN, L_BLK,
                        obj_list, R_BLK> {};

struct bool_intersection
    : seq<string<'i', 'n', 't', 'e', 'r', 's', 'e', 'c', 't', 'i', 'o', 'n'>,
          L_FUN, R_FUN, L_BLK, obj_list, R_BLK> {};

struct bool_diff : seq<string<'d', 'i', 'f', 'f', 'e', 'r', 'e', 'n', 'c', 'e'>,
                       L_FUN, R_FUN, L_BLK, obj_list, R_BLK> {};

struct bool_exp : sor<bool_union, bool_intersection, bool_diff> {};

struct mulmat : seq<string<'m', 'u', 'l', 't', 'm', 'a', 't', 'r', 'i', 'x'>,
                    L_FUN, matrix, R_FUN, L_BLK, obj_list, R_BLK> {};

struct group : seq<string<'g', 'r', 'o', 'u', 'p'>, L_FUN, R_FUN,
                   opt<seq<L_BLK, obj_list, R_BLK>>, opt<S_CLN>> {};

struct render : seq<string<'r', 'e', 'n', 'd', 'e', 'r'>, L_FUN, attr_list,
                    R_FUN, opt<seq<L_BLK, obj_list, R_BLK>>, opt<S_CLN>> {};

struct colorgroup : seq<string<'c', 'o', 'l', 'o', 'r'>, L_FUN, vector, R_FUN,
                        opt<seq<L_BLK, obj_list, R_BLK>>, opt<S_CLN>> {};

struct csg_obj : sor<shape, mulmat, bool_exp, group, render, colorgroup> {};

struct obj_list : plus<pad<csg_obj, space>> {};

struct grammar : seq<obj_list, eof> {};

template <typename Rule> struct action {};

template <> struct action<L_BLK> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    // std::cout << " { " << std::endl;
    std::vector<csg::Type> new_group;
    st.current_objs.push_back(new_group);
  }
};

template <> struct action<R_BLK> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    // std::cout << " } " << std::endl;
    st.current_group.clear();
    for (auto obj : st.current_objs.back()) {
      st.current_group.push_back(obj);
    }
    st.current_objs.pop_back();
  }
};

template <> struct action<vector_cell> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    std::stringstream ss(in.string());
    double cell;
    ss >> cell;
    st.current_vec.push_back(cell);
  }
};

template <> struct action<matrix> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    st.current_matrices.push_back(st.current_matrix);
    st.current_matrix.clear();
  }
};

template <> struct action<row> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    st.current_matrix.push_back(st.current_vec);
    st.current_vec.clear();
  }
};

void add_group(parser_state &st) {
  csg::Union group;
  for (const auto &curr_obj : st.current_group) {
    group.objs.push_back(curr_obj);
  }
  st.current_objs.back().push_back(group);
}

template <> struct action<group> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    add_group(st);
  }
};

template <> struct action<render> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    add_group(st);
  }
};

template <> struct action<colorgroup> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    add_group(st);
  }
};

template <> struct action<bool_union> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    add_group(st);
  }
};

template <> struct action<bool_intersection> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    auto csg_in = csg::Intersection();
    for (const auto &curr_obj : st.current_group) {
      csg_in.objs.push_back(curr_obj);
    }
    st.current_objs.back().push_back(csg_in);
  }
};

template <> struct action<bool_diff> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    auto csg_diff = csg::Difference();
    for (auto it = st.current_group.begin(); it != st.current_group.end();
         ++it) {
      if (it == st.current_group.begin()) {
        csg_diff.first_obj = std::make_shared<csg::Type>(*it);
      } else {
        csg_diff.next_objs.objs.push_back(*it);
      }
    }
    st.current_objs.back().push_back(csg_diff);
  }
};

template <> struct action<mulmat> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    auto mulmat = csg::Mulmatrix();
    auto mat = st.current_matrices.back();
    // clang-format off
    mulmat.rotation[0] = {mat[0][0], mat[0][1], mat[0][2]};
    mulmat.rotation[1] = {mat[1][0], mat[1][1], mat[1][2]};
    mulmat.rotation[2] = {mat[2][0], mat[2][1], mat[2][2]};
    // clang-format on

    mulmat.translation = {
        mat[0][3],
        mat[1][3],
        mat[2][3],
    };

    for (const auto &curr_obj : st.current_group) {
      mulmat.group.objs.push_back(curr_obj);
    }
    st.current_objs.back().push_back(mulmat);
    st.current_matrices.pop_back();
  }
};

template <> struct action<cube> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    csg::Cube cub;
    auto size = std::get<std::vector<double>>(st.curr_attr["size"]);
    cub.size = {size[0], size[1], size[2]};
    cub.center = std::get<bool>(st.curr_attr["center"]);

    st.current_objs.back().push_back(cub);
    st.curr_attr.clear();
  }
};

template <> struct action<cylinder> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    if (st.curr_attr.count("r")) {
      // proper cylinder
      if (st.curr_attr.count("r1") || st.curr_attr.count("r2")) {
        std::cout << " ERROR: cannot specify both r and (r1 or r2); ambiguous "
                  << std::endl;
      }
      csg::Cylinder cyl;
      cyl.center = std::get<bool>(st.curr_attr["center"]);
      cyl.height = std::get<double>(st.curr_attr["h"]);
      cyl.radius = std::get<double>(st.curr_attr["r"]);
      st.current_objs.back().push_back(cyl);

    } else {
      // conic "cylinder"
      csg::Cone cone;
      cone.center = std::get<bool>(st.curr_attr["center"]);
      cone.height = std::get<double>(st.curr_attr["h"]);
      cone.radius1 = std::get<double>(st.curr_attr["r1"]);
      cone.radius2 = std::get<double>(st.curr_attr["r2"]);
      st.current_objs.back().push_back(cone);
    }

    st.curr_attr.clear();
  }
};

template <> struct action<sphere> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    csg::Sphere sph;
    sph.radius = std::get<double>(st.curr_attr["r"]);

    st.current_objs.back().push_back(sph);
    st.curr_attr.clear();
  }
};

template <> struct action<name> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    std::stringstream ss(in.string());
    ss >> st.current_name;
  }
};

template <> struct action<string_literal> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    std::stringstream ss(in.string());
    std::string v;
    ss >> v;
    st.curr_attr[st.current_name] = v;
  }
};

template <> struct action<double_> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    std::stringstream ss(in.string());
    double v;
    ss >> v;
    st.curr_attr[st.current_name] = v;
  }
};

template <> struct action<true_literal> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    st.curr_attr[st.current_name] = true;
  }
};

template <> struct action<false_literal> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    st.curr_attr[st.current_name] = false;
  }
};

template <> struct action<vector_attr> {
  template <typename Input>
  static void apply(const Input &in, parser_state &st) {
    st.curr_attr[st.current_name] = st.current_vec;
    st.current_vec.clear();
  }
};

std::optional<parser_state> do_parse(std::string str) {
  parser_state st;
  std::vector<csg::Type> new_group;
  st.current_objs.push_back(new_group);
  memory_input in(str, "std::cin");
  if (!parse<grammar, action>(in, st)) {
    return std::nullopt;
  }
  return st;
}

} // namespace

namespace csg {

std::shared_ptr<Tree> parse_csg(std::string str) {
  auto maybe_state = do_parse(str);
  if (!maybe_state.has_value()) {
    return nullptr;
  }
  auto st = maybe_state.value();

  assert(st.current_objs.size() == 1);
  Tree tree;
  for (auto obj : st.current_objs.back()) {
    tree.top.objs.push_back(obj);
  }
  assert(tree.top.objs.size() > 0); // Disallow empty .csg file
  return std::make_shared<Tree>(tree);
}
} // namespace csg
